using System.Drawing;
using System.Drawing.Imaging;
using OSMDatastructure;
using OSMDatastructure.Graph;
using Pathfinding;
using RenderPath;
using Region = OSMDatastructure.Region;

namespace Server;

public class Server
{
    
    public static void Main(string[] args)
    {
        ConsoleWriter newConsole = new();
        Console.SetOut(newConsole);
        Console.SetError(newConsole);
        
        string workingDir = "D:/stuttgart-regbez-latest";

        //RegionConverter.ConvertXMLToRegions("D:/stuttgart-regbez-latest.osm", "D:/stuttgart-regbez-latest");
        //RegionConverter.ConvertXMLToRegions("D:/map.osm", "D:/map");
        //RegionConverter.ConvertXMLToRegions("D:/germany-latest.osm", "D:/germany-latest");
        
        Coordinates start = new (48.7933798f, 9.8275859f);
        Coordinates finish = new (48.795918f, 9.021618f);
        
        TestVariables(workingDir, start, finish, 16);
        GetShortestRoute("D:");

        /*
        ValueTuple<Image, Renderer.Bounds> area = Renderer.DrawArea(LoadRegions(workingDir, start, finish));
        area.Item1.Save(@"D:\Base.png", ImageFormat.Png);
        
        ValueTuple<Image, Renderer.Bounds> areaDistance = Renderer.DrawPath(
            PathResult.PathresultFromFile(@"D:\angle0,140_level0,020_same0,160.result"), Image.FromFile(@"D:\Base.png"), area.Item2);
        areaDistance.Item1.Save(@"D:\Distance.png", ImageFormat.Png);
        
        ValueTuple<Image, Renderer.Bounds> areaWeight = Renderer.DrawPath(
            PathResult.PathresultFromFile(@"D:\angle0,160_level0,020_same0,020.result"), Image.FromFile(@"D:\Base.png"), area.Item2);
        areaWeight.Item1.Save(@"D:\Weight.png", ImageFormat.Png);
*/
/*
        string parentFolder = new DirectoryInfo(workingDir).Parent!.FullName;
        
        Pathfinder result = new Pathfinder(workingDir, 0.002, 0,
            0, 1, 30).AStar(start,
            finish, Tag.SpeedType.car);

        Console.WriteLine($"Calc-time {result.pathResult!.calcTime} Path-length: {result.pathResult.pathNodes.Count} Visited-nodes: {result.gScore!.Count}");

        string fileName = DateTime.Now.ToFileTime().ToString();
        
        string resultFileName = $"{new DirectoryInfo(workingDir).Name}-{fileName}.result";
        result.SaveResult(Path.Join(parentFolder, resultFileName));

        string renderFileName = $"{new DirectoryInfo(workingDir).Name}-{fileName}.render.png";
        Image render = Renderer.DrawPathfinder(result);
#pragma warning disable CA1416
        render.Save(Path.Join(parentFolder, renderFileName), ImageFormat.Png);
#pragma warning restore CA1416*/
        Console.Beep(400, 100);
    }

    private static void GetShortestRoute(string directory)
    {
        DateTime start = DateTime.Now;
        HashSet<string> allFiles = Directory.GetFiles(directory).Where(file => file.EndsWith(".result")).ToHashSet();
        PathResult first = PathResult.PathresultFromFile(allFiles.First());
        KeyValuePair<PathResult, string> shortest = new(first, allFiles.First());
        KeyValuePair<PathResult, string> fastest = new(first, allFiles.First());
        KeyValuePair<PathResult, string> calcTime = new(first, allFiles.First());
        int loaded = 0;
        foreach (string filePath in allFiles)
        {
            PathResult result = PathResult.PathresultFromFile(filePath);
            Console.WriteLine($"{loaded++}/{allFiles.Count()} {filePath}  " +
                              $"Time elapsed: {DateTime.Now - start} " +
                              $"Remaining {((DateTime.Now - start)/loaded)*(allFiles.Count-loaded)}");
        }

        Dictionary<PathResult, string> sortedCalcTime = results.OrderBy(result => result.Key.calcTime).ToDictionary(item => item.Key, item =>item.Value);
        foreach (KeyValuePair<PathResult, string> kv in sortedCalcTime)
        {
            Console.WriteLine($"{kv.Key.distance:0.0} {kv.Key.weight:0.00} {kv.Key.calcTime} {kv.Value}");
        }

        KeyValuePair<PathResult, string> shortest = results.MinBy(result => result.Key.distance);
        KeyValuePair<PathResult, string> fastest = results.MinBy(result => result.Key.weight);
        KeyValuePair<PathResult, string> calcTime = results.MinBy(result => result.Key.calcTime);
        Console.WriteLine($"\nShortest:\t{shortest.Key.distance:0.0} {shortest.Key.weight:0.00} {shortest.Key.calcTime} {shortest.Value}\n" +
                          $"Fastest:\t{fastest.Key.distance:0.0} {fastest.Key.weight:0.00} {fastest.Key.calcTime} {fastest.Value}\n" +
                          $"CalcTime:\t{calcTime.Key.distance:0.0} {calcTime.Key.weight:0.00} {calcTime.Key.calcTime} {calcTime.Value}");
    }

    private static RegionManager LoadRegions(string workingDir, Coordinates c1, Coordinates c2)
    {
        float minLat = c1.latitude < c2.latitude ? c1.latitude : c2.latitude;
        float minLon = c1.longitude < c2.longitude ? c1.longitude : c2.longitude;
        float maxLat = c1.latitude > c2.latitude ? c1.latitude : c2.latitude;
        float maxLon = c1.longitude > c2.longitude ? c1.longitude : c2.longitude;

        RegionManager allRegions = new(workingDir);
        for (float lat = minLat - Region.RegionSize; lat < maxLat + Region.RegionSize; lat += Region.RegionSize / 2)
        {
            for (float lon = minLon - Region.RegionSize; lon < maxLon + Region.RegionSize; lon += Region.RegionSize / 2)
            {
                allRegions.GetRegion(new Coordinates(lat, lon));
            }
        }
        Console.WriteLine("Loaded needed Regions");
        return allRegions;
    }

    private static void TestVariables(string workingDir, Coordinates start, Coordinates finish, int threads)
    {
        string parentFolder = new DirectoryInfo(workingDir).Parent!.FullName;

        RegionManager rm = LoadRegions(workingDir, start, finish);
        
        Queue<Thread> calcThreads = new();

        for (double roadLevelPriority = 0.016; roadLevelPriority < 0.02; roadLevelPriority += 0.0002)
        {
            for (double maxAngle = 25; maxAngle < 35; maxAngle += 1)
            {
                double priority = roadLevelPriority;
                double angle = maxAngle;
                calcThreads.Enqueue(new Thread(() =>
                {
                    Pathfinder testresult = new Pathfinder(rm, priority, angle).AStar(start,
                        finish, Tag.SpeedType.car);
                    string fileName = $"angle{angle:00}_level{priority:0.0000}.result";
                    testresult.SaveResult(Path.Join(parentFolder, fileName));
                }));
            }
        }

        int totalTasks = calcThreads.Count;
        int completedTasks = 0;
        DateTime startTime = DateTime.Now;

        HashSet<Thread> runningThreads = new();
        while (calcThreads.Count > 0 || runningThreads.Count > 0)
        {
            while (runningThreads.Count < threads && calcThreads.Count > 0)
            {
                Thread t = calcThreads.Dequeue();
                runningThreads.Add(t);
                t.Start();
            }

            int newCompletedTasks = runningThreads.RemoveWhere(thread => !thread.IsAlive);
            completedTasks += newCompletedTasks;
            if (newCompletedTasks > 0)
            {
                TimeSpan elapsedTime = DateTime.Now - startTime;
                Console.WriteLine($"To calculate: {calcThreads.Count}(+{runningThreads.Count} running)/{totalTasks} Time Average: {(elapsedTime/completedTasks)} Elapsed: {elapsedTime} Remaining: {(elapsedTime/completedTasks*calcThreads.Count)}");
            }
        }
    }
}